@@ -40,11 +40,17 @@ typedef unsigned long long u64;
 #include "nbd.h"
 
 #if NBD_LFS==1
+/* /usr/include/features.h (included from /usr/include/sys/types.h)
+   defines this when _GNU_SOURCE is defined
+ */
+#ifndef _LARGEFILE_SOURCE
 #define _LARGEFILE_SOURCE
+#endif
 #define _FILE_OFFSET_BITS 64
 #endif
 
 u64 cliserv_magic = 0x00420281861253LL;
+u64 opts_magic = 0x49484156454F5054LL;
 #define INIT_PASSWD "NBDMAGIC"
 
 #define INFO(a) do { } while(0)
@@ -70,18 +76,17 @@ void setmysockopt(int sock) {
 #ifndef G_GNUC_NORETURN
 #if __GNUC__ > 2 || (__GNUC__ == 2 && __GNUC_MINOR__ > 4)
 #define G_GNUC_NORETURN __attribute__((__noreturn__))
+#define G_GNUC_UNUSED __attribute__((unused))
 #else
 #define G_GNUC_NORETURN
+#define G_GNUC_UNUSED
 #endif
 #endif
 
-void err(const char *s) G_GNUC_NORETURN;
+void err_nonfatal(const char *s) {
+	char s1[150], *s2;
 
-void err(const char *s) {
-	const int maxlen = 150;
-	char s1[maxlen], *s2;
-
-	strncpy(s1, s, maxlen);
+	strncpy(s1, s, sizeof(s1));
 	if ((s2 = strstr(s, "%m"))) {
 		strcpy(s1 + (s2 - s), strerror(errno));
 		s2 += 2;
@@ -96,13 +101,19 @@ void err(const char *s) {
 	}
 #endif
 
-	s1[maxlen-1] = '\0';
+	s1[sizeof(s1)-1] = '\0';
 #ifdef ISSERVER
 	syslog(LOG_ERR, "%s", s1);
 	syslog(LOG_ERR, "Exiting.");
 #endif
 	fprintf(stderr, "Error: %s\nExiting.\n", s1);
-	exit(1);
+}
+
+void err(const char *s) G_GNUC_NORETURN;
+
+void err(const char *s) {
+	err_nonfatal(s);
+	exit(EXIT_FAILURE);
 }
 
 void logging(void) {
@@ -128,6 +139,8 @@ u64 ntohll(u64 a) {
 #endif
 #define htonll ntohll
 
-/* Flags used between the client and server */
-#define NBD_FLAG_HAS_FLAGS	(1 << 0)	/* Flags are there */
-#define NBD_FLAG_READ_ONLY	(1 << 1)	/* Device is read-only */
+#define NBD_DEFAULT_PORT	"10809"	/* Port on which named exports are
+					 * served */
+
+/* Options that the client can select to the server */
+#define NBD_OPT_EXPORT_NAME	(1 << 0)	/* Client wants to select a named export (is followed by length and name of export) */
